
#ifndef ATLAS_TYPES_H
#define ATLAS_TYPES_H


#include "../Atlas.h" // must be very first \.{atlas} include
#include "axis-types.h"

#include <stdexcept>
#include "lietype.h"

#include "rootdata.h"

#include "innerclass.h"

#include "output.h"

#include "realredgp.h"

#include "cartanclass.h"

#include "blocks.h"

#include "repr.h"

#include "arithmetic.h"

#include "free_abelian.h" 

namespace atlas { namespace interpreter {

struct Lie_type_value : public value_base
{ LieType val;

  Lie_type_value() : val()  {}
    // default constructor, produces empty type
  Lie_type_value(LieType t) : val(t) {}
    // constructor from already validated Lie type

  virtual void print(std::ostream& out) const;
  Lie_type_value* clone() const { return new Lie_type_value(*this); }
  static const char* name() { return "Lie type"; }

  void add_simple_factor (char,size_t); // grow
private:
  Lie_type_value(const Lie_type_value& v) : val(v.val) {}
    // copy constructor, used by |clone|
};

typedef std::shared_ptr<const Lie_type_value> shared_Lie_type;
typedef std::shared_ptr<Lie_type_value> own_Lie_type;

struct root_datum_value : public value_base
{ RootDatum val;

  root_datum_value(const RootDatum& v) : val(v)  {}
  virtual void print(std::ostream& out) const;
  root_datum_value* clone() const { return new root_datum_value(*this); }
  static const char* name() { return "root datum"; }
private:
  root_datum_value(const root_datum_value& v) : val(v.val) {}
};

typedef std::unique_ptr<root_datum_value> root_datum_ptr;
typedef std::shared_ptr<const root_datum_value> shared_root_datum;

struct inner_class_value : public value_base
{ InnerClass& val;
  InnerClass& dual;
  size_t& ref_count;

  lietype::LieType rd_type;
  lietype::InnerClassType ic_type;
  const output::FormNumberMap interface,dual_interface;

  inner_class_value  // main constructor
   (std::unique_ptr<InnerClass> G, const lietype::Layout& lo);
  ~inner_class_value();

  virtual void print(std::ostream& out) const;
  inner_class_value* clone() const 
    { return new inner_class_value(*this); }
  static const char* name() { return "inner class"; }

  inner_class_value(const inner_class_value& v); // copy constructor
  inner_class_value(const inner_class_value& v,tags::DualTag);
   // constructor of dual
};

typedef std::shared_ptr<const inner_class_value> shared_inner_class;

struct real_form_value : public value_base
{ const inner_class_value parent;
  RealReductiveGroup val;

  real_form_value(const inner_class_value& p,RealFormNbr f) 
  : parent(p), val(p.val,f)
  , khc_p(nullptr)
  , rt_p(nullptr) {}
  real_form_value
    (const inner_class_value& p,RealFormNbr f
    ,const RatCoweight& coch, TorusPart tp) 
  : parent(p), val(p.val,f,coch,tp)
  , khc_p(nullptr)
  , rt_p(nullptr) {}
  virtual ~real_form_value ();

  virtual void print(std::ostream& out) const;
  real_form_value* clone() const 
    { return new real_form_value(*this); }
  static const char* name() { return "real form"; }
  const KGB& kgb () { return val.kgb(); }
   // generate and return $K\backslash G/B$ set
  KhatContext& khc();
  const Rep_context& rc();
  Rep_table& rt();
private:
  KhatContext* khc_p;
  Rep_table* rt_p;
    // owned pointers, initially |nullptr|, assigned at most once
};

typedef std::shared_ptr<const real_form_value> shared_real_form;
typedef std::shared_ptr<real_form_value> own_real_form;

struct Cartan_class_value : public value_base
{ const inner_class_value parent;
  size_t number;
  const CartanClass& val;

  Cartan_class_value(const inner_class_value& p,size_t cn);
  ~Cartan_class_value() {} // everything is handled by destructor of |parent|

  virtual void print(std::ostream& out) const;
  Cartan_class_value* clone() const 
    { return new Cartan_class_value(*this); }
  static const char* name() { return "Cartan class"; }
private:
  Cartan_class_value(const Cartan_class_value& v)
  : parent(v.parent), number(v.number), val(v.val) {} // copy constructor
};

typedef std::shared_ptr<const Cartan_class_value> shared_Cartan_class;

struct KGB_elt_value : public value_base
{ own_real_form rf;
  KGBElt val;

  KGB_elt_value(const own_real_form& form, KGBElt x) : rf(form), val(x) {}
  ~KGB_elt_value() {}

  virtual void print(std::ostream& out) const;
  KGB_elt_value* clone() const  { return new KGB_elt_value(*this); }
  static const char* name() { return "KGB element"; }
private:
  KGB_elt_value(const KGB_elt_value& v)
  : rf(v.rf), val(v.val) {} // copy constructor
};

typedef std::unique_ptr<KGB_elt_value> KGB_elt_ptr;
typedef std::shared_ptr<const KGB_elt_value> shared_KGB_elt;
typedef std::shared_ptr<KGB_elt_value> own_KGB_elt;

struct Block_value : public value_base
{ const own_real_form rf; const own_real_form dual_rf;
  Block val; // cannot be |const|, as Bruhat order may be generated implicitly

  Block_value(const own_real_form& form, const own_real_form& dual_form);
  ~Block_value() {}

  virtual void print(std::ostream& out) const;
  Block_value* clone() const  { return new Block_value(*this); }
  static const char* name() { return "KGB element"; }
private:
  Block_value(const Block_value& v)
  : rf(v.rf), val(v.val) {} // copy constructor
};

typedef std::unique_ptr<Block_value> Block_ptr;
typedef std::shared_ptr<const Block_value> shared_Block;
typedef std::shared_ptr<Block_value> own_Block;

struct module_parameter_value : public value_base
{ own_real_form rf;
  StandardRepr val;

  module_parameter_value(const own_real_form& form, const StandardRepr& v)
  : rf(form), val(v) {}
  ~module_parameter_value() {}

  virtual void print(std::ostream& out) const;
  module_parameter_value* clone() const
    { return new module_parameter_value(*this); }
  static const char* name() { return "module parameter"; }

  const Rep_context& rc() const { return rf->rc(); }
  Rep_table& rt() const { return rf->rt(); }
private:
  module_parameter_value(const module_parameter_value& v)  // copy constructor
  : rf(v.rf),val(v.val)  {}
};

typedef std::unique_ptr<module_parameter_value> module_parameter_ptr;
typedef std::shared_ptr<const module_parameter_value> shared_module_parameter;
typedef std::shared_ptr<module_parameter_value> own_module_parameter;

struct split_int_value : public value_base
{ Split_integer val;

  explicit split_int_value(Split_integer v) : val(v)  {}
  ~split_int_value() {}
  void print(std::ostream& out) const;
  split_int_value* clone() const { return new split_int_value(*this); }
  static const char* name() { return "split integer"; }
private:
  split_int_value(const split_int_value& v) : val(v.val) {}
};

typedef std::unique_ptr<split_int_value> split_int_ptr;
typedef std::shared_ptr<const split_int_value> shared_split_int;
typedef std::shared_ptr<split_int_value> own_split_int;

struct virtual_module_value : public value_base
{ own_real_form rf;
  repr::SR_poly val;

  virtual_module_value(const own_real_form& form, const repr::SR_poly& v)
  : rf(form), val(v) {}
  ~virtual_module_value() {}

  virtual void print(std::ostream& out) const;
  virtual_module_value* clone() const
    { return new virtual_module_value(*this); }
  static const char* name() { return "module parameter"; }

  const Rep_context& rc() const { return rf->rc(); }
  Rep_table& rt() const { return rf->rt(); }
private:
  virtual_module_value(const virtual_module_value& v)
   : rf(v.rf),val(v.val) {} // copy
};

typedef std::unique_ptr<virtual_module_value> virtual_module_ptr;
typedef std::shared_ptr<const virtual_module_value> shared_virtual_module;
typedef std::shared_ptr<virtual_module_value> own_virtual_module;


void initialise_builtin_types();

} }
#endif

