
#include "lexer.h"
namespace atlas
{ namespace interpreter
   {

Lexical_analyser::Lexical_analyser
  (BufferedInput& source, Hash_table& hash,
   const char** keywords, const char** type_names)
: input(source),id_table(hash),nesting(0)
 ,prevent_termination('\0'),state(initial)
{ 
  { for (size_t i=0; keywords[i]!=0; ++i)
      id_table.match_literal(keywords[i]);
    keyword_limit=id_table.nr_entries();
    for (size_t i=0; type_names[i]!=0; ++i)
      id_table.match_literal(type_names[i]);
    type_limit=id_table.nr_entries();
  }
  comment_start=comment_end=256; // a non-|char| value
}

bool Lexical_analyser::reset()
{ nesting=0; state=initial; input.reset(); return input.getline(); }

void Lexical_analyser::skip_space() const
{ if (prevent_termination!='\0') input.push_prompt(prevent_termination);
  do
  { char c=input.shift();
    if (std::isspace(c))
     // ignore unless file ends, or a newline where a command could end
  { if (c=='\f' or
            c=='\n' and prevent_termination=='\0' and nesting==0)
        break;
    }
    else if (c==comment_start) 
                             { int line, column;
                               input.locate(input.point(),line,column); // maybe needed for reporting
                               if (comment_end=='\n' or comment_end==comment_start)
                                  // non nestable comment
                               { do c=input.shift();
                               while (c!=comment_end and c!='\0' and c!='\f');
                                 if (c!=comment_end)
                                 { std::cerr << "Comment that started on line " << line
                                             << ", column " << column  << " is never closed.\n";
                                 input.unshift(); // reconsider end of file
                                 }
                                 else if (comment_end=='\n')
                                   input.unshift(); // reconsider newline character
                               }
                               else // skip comment, allowing for nested comment groups
                               { int level=1;
                                 input.push_prompt(comment_start);
                                 do
                                 {
                                   do c=input.shift();  
                                   while (c!= comment_start and c!=comment_end and c!='\0' and c!='\f');
                                   if (c=='\0' or c=='\f')
                                   { std::cerr << "Comment that started on line " << line
                                               << ", column " << column  << " is never closed.\n";
                                    input.unshift(); break;
                                   // force out of comment loop, reconsider character
                                   }
                                   if (c==comment_start)
                                   { ++level;
                                     input.push_prompt(comment_start);
                                   }
                                   else
                                   { --level;
                                     input.pop_prompt();
                                   } // |c==comment_end|
                                 }
                                 while (level>0);
                               }
                             } 
    else break;
      // non-space and non-comment character (possibly |'\0'|): terminate
  } while(true);
  if (prevent_termination!='\0') input.pop_prompt();
  input.unshift(); // prepare to re-read character that ended space
}

std::string Lexical_analyser::scan_quoted_string() const
{ const char* start=input.point(); bool broken=false;
  std::string result; char c;
  while (true)
  { for (c=input.shift(); c!='"' and c!='\n' and c!='\0'; c=input.shift())
      result.push_back(c);
    if (c!='"')
      { broken=true; break; }
    else if ((c=input.shift())!='"')
      break; // normal ending of string
    result.push_back(c); // doubled quote; insert one copy and continue
  }
  input.unshift();
  if (broken)
  { const char* end=input.point();
    int l0,c0,l1,c1;
    input.locate(start,l0,c0); input.locate(end,l1,c1);
    input.show_range(std::cerr,l0,c0,l1,c1);
    std::cerr << "Closing string denotation.\n";
  }
  return result;
}

int Lexical_analyser::get_token(YYSTYPE *valp, YYLTYPE* locp)
{ if (state==ended)
    { state=initial; return 0; } // send end of input
  skip_space(); prevent_termination='\0';
  input.locate(input.point(),locp->first_line,locp->first_column);
  int code; char c=input.shift();
  if (std::isalpha(c) or c=='_')
    
    { const char* p=input.point()-1; // start of token
      do
        c=input.shift();
      while(std::isalnum(c) || c=='_');
      input.unshift();
      id_type id_code=id_table.match(p,input.point()-p);
      if (id_code>=type_limit)
      { valp->id_code=id_code;
        if (global_id_table->is_defined_type(id_code))
          code=TYPE_ID;
        else
          code=IDENT;
      }
      else if (id_code>=keyword_limit)
      { valp->type_code=id_code-keyword_limit; code=TYPE; }
      else // we have |id_code<keyword_limit|, so it is a keyword
      { code=QUIT+id_code;
        switch(code)
        {
          case LET: ++nesting; input.push_prompt('L'); break;
          case BEGIN:
          case IF:
          case WHILE:
          case FOR: ++nesting; input.push_prompt('G'); break;
          case IN: if (input.top_prompt()=='L')
          { --nesting; input.pop_prompt(); prevent_termination='I'; }
          break;
          case END:
          case FI:
          case OD: --nesting; input.pop_prompt(); break;
          case AND: case OR: case NOT: prevent_termination='~'; break;
          case WHATTYPE: prevent_termination='W'; break;
        }
      }
    }
  else if (std::isdigit(c))
    
    { const char* p=input.point()-1; // start of token
      do
        c=input.shift();
      while(std::isdigit(c));
      input.unshift();
      const char* end=input.point();
      unsigned long val=*p++-'0';   while (p<end) val=10*val+(*p++-'0');
      valp->val=val; code=INT;
    }
  else
    
    { switch(c)
      {      case '"': 
                    { valp->str = new std::string(scan_quoted_string()); code=STRING; }
     
      break; case '(':
             case '{':
             case '[': ++nesting; input.push_prompt(c); code=c;
      break; case ')':
             case '}':
             case ']': --nesting; input.pop_prompt(); code=c;
      break; case ';': code=prevent_termination=c;
      break; case '<':
             case '>':
             if (state==initial)
             { code= c=='<'
               ? input.shift()=='<' ? FORCEFROMFILE:  (input.unshift(),FROMFILE)
               : input.shift()=='>' ? ADDTOFILE : (input.unshift(),TOFILE) ;
    	   
    	   if ((skip_space(),c=input.shift())=='"')
    	     file_name=scan_quoted_string();
    	   else
    	   { file_name="";
    	     while (std::isalnum(c) or std::strchr(".-+~_=!?@#$%&|",c)!=nullptr)
    	     { file_name.push_back(c); c=input.shift(); }
    	     input.unshift();
    	   }
                break;
             }
             prevent_termination=c;
             valp->oper.priority=2;
             if (input.shift()=='=')
               valp->oper.id=id_table.match_literal(c=='<' ? "<=" : ">=");
             else
               { input.unshift();
                   valp->oper.id=id_table.match_literal(c=='<' ? "<" : ">");
               }
             code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
      break; case ':': prevent_termination=c;
        code = input.shift()=='=' ? BECOMES  : (input.unshift(),c);
      break; case '=':
             valp->oper.id = id_table.match_literal("=");
             valp->oper.priority = 2; // in case
    	 prevent_termination=c;
    	 code = becomes_follows() ? OPERATOR_BECOMES : '=';
      break; case '!':
             if (input.shift()=='=')
             { valp->oper.id = id_table.match_literal("!=");
               valp->oper.priority = 2;
               prevent_termination='=';
               code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
             }
             else
               code= (input.unshift(),c);
             // currently unused; might some day be factorial operator
      break; case '~': 
                    if (input.shift()=='[') // recognise combination for parse reason
                    { code = TLSUB; ++nesting; input.push_prompt('['); }
                    else
                     // now see if, skipping spaces and comments, next can follow |'~'|
                    { input.unshift(); prevent_termination='~'; skip_space();
                       // for these cases we do allow intervening space, comments
                      const char* p = input.point();
                       // take a peek into the input buffer without advancing
                      if (*p==':' or *p==']' or *p==',' or
                       (*p=='d' and p[1]=='o' or *p=='o' and p[1]=='d')
                       and  not (p[2]=='_' or std::isalnum((unsigned char)p[2])))
                       code=c; // return |'~'|
                      else
                      { valp->oper.id = id_table.match_literal("~");
                        valp->oper.priority = 8; // same precedence as \.{\#}
                        code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
                      }
                    }
      break; case '\f': code=END_OF_FILE; // tell the parser a file ended
      break; 
                 case '+': prevent_termination=code=c;
                    valp->oper.id = id_table.match_literal("+");
                    valp->oper.priority = 4;
                    code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
             break; case '-': prevent_termination=c;
                    if (input.shift()=='>')
                       code= ARROW;
                    else
                    { input.unshift();
                      valp->oper.id = id_table.match_literal("-");
                      valp->oper.priority = 4;
                      code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
                    }
             break; case '*': prevent_termination=c;
                    valp->oper.id =
                       id_table.match_literal(c=='*' ? "*" : c=='%' ? "%" : "/");
                    valp->oper.priority = 6;
                    code = becomes_follows() ? OPERATOR_BECOMES : c;
             break; case '%': case '/': prevent_termination=c;
                    valp->oper.id =
                       id_table.match_literal(c=='%' ? "%" : "/");
                    valp->oper.priority = 6;
                    code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
             break; case '\\':
                    prevent_termination=c;
                    valp->oper.priority = 6;
                    if (input.shift()=='%')
                    { prevent_termination='%';
                      valp->oper.id = id_table.match_literal("\\%");
                    }
                    else
                    { input.unshift();
                      valp->oper.id = id_table.match_literal("\\");
                    }
                    code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
             break; case '^': prevent_termination=c;
                    valp->oper.id = id_table.match_literal("^");
                    valp->oper.priority = 7; // exponentiation is right-associative
                    code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
             break; case '#': prevent_termination=c;
                    valp->oper.priority = 8;
                    if (input.shift()=='#')
                      valp->oper.id = id_table.match_literal("##");
                    else
                    { input.unshift();
                      valp->oper.id = id_table.match_literal("#");
                    }
                    code = becomes_follows() ? OPERATOR_BECOMES : OPERATOR;
             break;
             case '\n': state=ended; // and {\bf fall through}.
             default: code=c;
      }
    }
  input.locate(input.point(),locp->last_line,locp->last_column);
if (state==initial)
   state=normal;
  return code;
}

bool Lexical_analyser::becomes_follows ()
{ skip_space(); // allow intervening space, comments
  const char* p = input.point();
   // take a peek into the input buffer without advancing
  if (*p==':' and *++p=='=') // then indeed |":="| follows our operator
  { input.shift(); input.shift(); // gobble up the symbol |":="|
    prevent_termination = ':'; return true;
  }
  return false;
}


Hash_table* main_hash_table=nullptr;

Lexical_analyser* lex=nullptr;

   }
}

